@@ -1,13 +1,25 @@
 import copy
 import os
 import logging
-import archinfo
+from typing import List, Optional, Dict
+from collections import OrderedDict, defaultdict
+from sortedcontainers import SortedDict
+
 import elftools
 from elftools.elf import elffile, sections
 from elftools.dwarf import callframe
 from elftools.common.exceptions import ELFError, DWARFError
-from collections import OrderedDict, defaultdict
+from elftools.dwarf.dwarf_expr import DWARFExprParser
+from elftools.dwarf.descriptions import describe_form_class, describe_attr_value
+from elftools.dwarf.dwarfinfo import DWARFInfo
+from elftools.dwarf.die import DIE
 
+import archinfo
+
+from .compilation_unit import CompilationUnit
+from .variable import Variable
+from .variable_type import VariableType
+from .subprogram import Subprogram
 from .symbol import ELFSymbol, Symbol, SymbolType
 from .regions import ELFSection, ELFSegment
 from .hashtable import ELFHashTable, GNUHashTable
@@ -38,11 +50,11 @@ class ELF(MetaELF):
         try:
             self._reader = elffile.ELFFile(self._binary_stream)
             list(self._reader.iter_sections())
-        except Exception: # pylint: disable=broad-except
+        except Exception as e: # pylint: disable=broad-except
             self._binary_stream.seek(4)
             ty = self._binary_stream.read(1)
             if ty not in (b'\1', b'\2'):
-                raise CLECompatibilityError
+                raise CLECompatibilityError from e
 
             if ty == b'\1':
                 patch_data = [(0x20, b'\0'*4), (0x2e, b'\0'*6)]
@@ -65,9 +77,12 @@ class ELF(MetaELF):
         if self.arch is None:
             self.set_arch(self.extract_arch(self._reader))
         else:
-            other_arch = self.extract_arch(self._reader)
-            if other_arch != self.arch:
-                l.warning("User specified %s but autodetected %s. Proceed with caution.", self.arch, other_arch)
+            try:
+                other_arch = self.extract_arch(self._reader)
+                if other_arch != self.arch:
+                    l.warning("User specified %s but autodetected %s. Proceed with caution.", self.arch, other_arch)
+            except archinfo.ArchNotFound:
+                pass
 
         self._addend = addend
 
@@ -101,6 +116,9 @@ class ELF(MetaELF):
         #
         self.has_dwarf_info = bool(self._reader.has_dwarf_info())
         self.build_id = None
+        self.addr_to_line = SortedDict()
+        self.variables: Optional[List[Variable]] = None
+        self.compilation_units: Optional[List[CompilationUnit]] = None
 
         self._entry = self._reader.header.e_entry
         self.is_relocatable = self._reader.header.e_type == 'ET_REL'
@@ -136,14 +154,23 @@ class ELF(MetaELF):
                 dwarf = None
                 self.has_dwarf_info = False
 
-            if dwarf and dwarf.has_EH_CFI():
-                self._load_function_hints_from_fde(dwarf, FunctionHintSource.EH_FRAME)
-                self._load_exception_handling(dwarf)
+            if dwarf:
+                # Load DIEs
+                self._load_dies(dwarf)
+                # Load function hints and exception handling artifacts
+                if dwarf.has_EH_CFI():
+                    self._load_function_hints_from_fde(dwarf, FunctionHintSource.EH_FRAME)
+                    self._load_exception_handling(dwarf)
+                    self._load_line_info(dwarf)
 
         if debug_symbols:
             self.__process_debug_file(debug_symbols)
-        elif self.loader._load_debug_info and self.build_id:
-            debug_filename = '/usr/lib/debug/.build-id/%s/%s.debug' % (self.build_id[:2], self.build_id[2:])
+        elif self.loader._load_debug_info:
+            if self.build_id:
+                debug_filename = '/usr/lib/debug/.build-id/%s/%s.debug' % (self.build_id[:2], self.build_id[2:])
+                if os.path.isfile(debug_filename):
+                    self.__process_debug_file(debug_filename)
+            debug_filename = os.path.join('/usr/lib/debug', os.path.realpath(self.binary))
             if os.path.isfile(debug_filename):
                 self.__process_debug_file(debug_filename)
 
@@ -318,6 +345,12 @@ class ELF(MetaELF):
         else:
             raise CLEError("Bad symbol identifier: %r" % (symid,))
 
+    def rebase(self, new_base):
+        delta = new_base - self.linked_base
+        super().rebase(new_base)
+
+        self.addr_to_line = {addr + delta: value for addr, value in self.addr_to_line.items()}
+
     #
     # Private Methods
     #
@@ -492,6 +525,171 @@ class ELF(MetaELF):
             l.warning("An exception occurred in pyelftools when loading FDE information.",
                       exc_info=True)
 
+    def _load_line_info(self, dwarf):
+        """
+        Generates addr_to_line as a mapping: addr -> (filename, lineno).
+        Lineno is one-indexed.
+        """
+        for cu in dwarf.iter_CUs():
+            comp_dir = '.'
+            die = cu.get_top_DIE()
+            if 'DW_AT_comp_dir' in die.attributes:
+                comp_dir = die.attributes['DW_AT_comp_dir'].value
+            lineprog = dwarf.line_program_for_CU(cu)
+            if lineprog is None:
+                continue
+            file_cache = {}
+            for line in lineprog.get_entries():
+                if line.state is None:
+                    continue
+                if line.state.file in file_cache:
+                    filename = file_cache[line.state.file]
+                else:
+                    file_entry = lineprog.header['file_entry'][line.state.file - 1]
+                    if file_entry["dir_index"] == 0:
+                        filename = os.path.join(comp_dir, file_entry.name).decode()
+                    else:
+                        filename = os.path.join(
+                            comp_dir,
+                            lineprog.header["include_directory"][file_entry["dir_index"] - 1],
+                            file_entry.name).decode()
+                    file_cache[line.state.file] = filename
+
+                relocated_addr = AT.from_lva(line.state.address, self).to_mva()
+                self.addr_to_line[relocated_addr] = (filename, line.state.line)
+
+    @staticmethod
+    def _load_low_high_pc_form_die(die: DIE):
+        """
+        Load low and high pc from a DIE.
+
+        :param die:     The DIE object from pyelftools.
+        :return:        low_pc, high_pc
+        """
+        if 'DW_AT_low_pc' not in die.attributes:
+            return None, None
+        lowpc = die.attributes['DW_AT_low_pc'].value
+
+        if 'DW_AT_high_pc' not in die.attributes:
+            return lowpc, None
+
+        # DWARF v4 in section 2.17 describes how to interpret the
+        # DW_AT_high_pc attribute based on the class of its form.
+        # For class 'address' it's taken as an absolute address
+        # (similarly to DW_AT_low_pc); for class 'constant', it's
+        # an offset from DW_AT_low_pc.
+        highpc_attr = die.attributes['DW_AT_high_pc']
+        highpc_attr_class = describe_form_class(highpc_attr.form)
+        if highpc_attr_class == 'address':
+            highpc = highpc_attr.value
+        elif highpc_attr_class == 'constant':
+            highpc = lowpc + highpc_attr.value
+        else:
+            l.warning('Error: invalid DW_AT_high_pc class:%s',highpc_attr_class)
+            return lowpc, None
+        return lowpc,highpc
+
+    def _load_dies(self, dwarf: DWARFInfo):
+        """
+        Load DIEs and CUs from DWARF.
+
+        :param dwarf:   The DWARF info object from pyelftools.
+        :return:        None
+        """
+        compilation_units: List[CompilationUnit] = [ ]
+        type_list: Dict[int, VariableType] = {}
+
+        for cu in dwarf.iter_CUs():
+            expr_parser = DWARFExprParser(cu.structs)
+
+            # scan the whole die tree for DW_TAG_base_type
+            for die in cu.iter_DIEs():
+                if die.tag == "DW_TAG_base_type":
+                    type_list[die.offset] = VariableType.read_from_die(die)
+
+            top_die = cu.get_top_DIE()
+
+            if top_die.tag != "DW_TAG_compile_unit":
+                l.warning("ignore a top die with unexpected tag")
+                continue
+
+            die_name = top_die.attributes['DW_AT_name'].value.decode('utf-8')
+            die_comp_dir = top_die.attributes['DW_AT_comp_dir'].value.decode('utf-8')
+            die_low_pc, die_high_pc = self._load_low_high_pc_form_die(top_die)
+            die_lang = describe_attr_value(top_die.attributes['DW_AT_language'], top_die, top_die.offset)
+
+            cu_ = CompilationUnit(die_name,die_comp_dir, die_low_pc, die_high_pc,die_lang)
+            compilation_units.append(cu_)
+
+            for die_child in cu.iter_DIE_children(top_die):
+                if die_child.tag == 'DW_TAG_variable':
+                    # load global variable
+                    var = self._load_die_variable(die_child, expr_parser, type_list)
+                    var.decl_file = cu_.file_path
+                    cu_.global_variables.append(var)
+                elif die_child.tag == 'DW_TAG_subprogram':
+                    # load subprogram
+                    
+                    if 'DW_AT_name' in die_child.attributes:
+                        name = die_child.attributes['DW_AT_name'].value.decode('utf-8')
+                    else:
+                        name = None
+                    low_pc, high_pc = self._load_low_high_pc_form_die(die_child)
+                    if low_pc is not None or high_pc is not None:
+                        sub_prog = Subprogram(name, low_pc, high_pc)
+
+                        for sub_die in cu._iter_DIE_subtree(die_child):
+                            if sub_die.tag in ['DW_TAG_variable','DW_TAG_formal_parameter']:
+                                # load local variable
+                                var = self._load_die_variable(sub_die, expr_parser, type_list)
+                                var.decl_file = cu_.file_path
+                                sub_prog.local_variables.append(var)
+
+                        cu_.functions[low_pc] = sub_prog
+
+        self.compilation_units = compilation_units
+
+    @staticmethod
+    def _load_die_variable(die: DIE, expr_parser, type_list) -> Variable:
+
+        if 'DW_AT_name' in die.attributes:
+            var_name = die.attributes['DW_AT_name'].value.decode('utf-8')
+        else:
+            var_name = None
+
+        if 'DW_AT_decl_line' in die.attributes:
+            decl_line = die.attributes['DW_AT_decl_line'].value
+        else:
+            decl_line = None
+
+        type_ = None
+        if 'DW_AT_type' in die.attributes:
+            type_offset = die.attributes['DW_AT_type'].value
+            if type_offset in type_list:
+                type_ = type_list[type_offset]
+            else:
+                # this warning takes too long to print on binaries with too many unknown type offsets. disable it
+                # l.warning("unknown type offset var_name:%s type_offset:%s", var_name, type_offset)
+                pass
+
+        v = Variable(
+            name= var_name,
+            type_= type_,
+            decl_file= None,  # decl_file: will be back-patched in _load_dies()
+            decl_line= decl_line,
+        )
+
+        if 'DW_AT_location' in die.attributes and die.attributes['DW_AT_location'].form == 'DW_FORM_exprloc':
+            parsed_exprs = expr_parser.parse_expr(die.attributes['DW_AT_location'].value)
+            if len(parsed_exprs) == 1 and parsed_exprs[0].op_name == 'DW_OP_addr':
+                v.sort, v.addr = "global", parsed_exprs[0].args[0]
+            elif len(parsed_exprs) == 1 and parsed_exprs[0].op_name == 'DW_OP_fbreg':
+                v.sort, v.addr = "stack", parsed_exprs[0].args[0]
+            elif len(parsed_exprs) == 1 and parsed_exprs[0].op_name.startswith("DW_OP_reg"):
+                v.sort, v.addr = "register", parsed_exprs[0].op - 0x50 # 0x50 == DW_OP_reg0
+
+        return v
+
     #
     # Private Methods... really. Calling these out of context
     # will probably break things. Caveat emptor.
@@ -547,6 +745,9 @@ class ELF(MetaELF):
         self.extra_load_path = self.__parse_rpath(runpath, rpath)
 
         strtab = seg_readelf._get_stringtable()
+        if strtab is None:
+            l.warning("Unexpected return value from pyelftools: stringtable object is None.")
+            return
         self.__neuter_streams(strtab)
 
         # To extract symbols from binaries without section headers, we need to hack into pyelftools.
@@ -742,10 +943,11 @@ class ELF(MetaELF):
                 self.jmprel[reloc.symbol.name] = reloc
 
     def __register_tls(self, seg_readelf):
-        self.tls_used = True
         self.tls_block_size = seg_readelf.header.p_memsz
         self.tls_data_size = seg_readelf.header.p_filesz
         self.tls_data_start = AT.from_lva(seg_readelf.header.p_vaddr, self).to_rva()
+        if self.tls_block_size != 0 or self.tls_data_size != 0:
+            self.tls_used = True
 
     def __register_relro(self, segment_relro):
         segment_relro = ELFSegment(segment_relro, relro=True)
@@ -937,18 +1139,18 @@ class ELF(MetaELF):
                 elif sec_readelf.header.sh_type == 'SHT_NOTE':
                     self.__register_notes(sec_readelf)
 
-            #has_dwarf_info = bool(elf.has_dwarf_info())
-            #if has_dwarf_info:
-            #    try:
-            #        dwarf = self.reader.get_dwarf_info()
-            #    except ELFError:
-            #        l.warning("An exception occurred in pyelftools when loading the DWARF information on %s.", filename,
-            #                  exc_info=True)
-            #        dwarf = None
+            has_dwarf_info = bool(elf.has_dwarf_info())
+            if has_dwarf_info:
+                try:
+                    dwarf = elf.get_dwarf_info()
+                except ELFError:
+                    l.warning("An exception occurred in pyelftools when loading the DWARF information on %s.", filename,
+                              exc_info=True)
+                    dwarf = None
 
                 # debug symbols don't have eh_frame ever from what I can tell
-                #if dwarf and dwarf.has_EH_CFI():
-                #    self._load_function_hints_from_fde(dwarf, FunctionHintSource.EXTERNAL_EH_FRAME)
+                if dwarf:
+                    self._load_line_info(dwarf)
 
 
 register_backend('elf', ELF)
